/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|          Program Name: Antennenumschalter                                                                             |
|                Author: Michael Krmer                                                                                 |
|                 Email: Michael@dk8pp.de                                                                               |
|                  Date: 26-Nov-2020                                                                                    |
|                  Tool: AVR Studio 7.0                                                                                 |
|                Device: ATMega644PA-AU                                                                                 |
|               Version: 0.1                                                                                            |
|        Program status: still testing                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: Antenna switch                                                                                              |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+

Preprocessor-Macros. Define the following preprocessor symbols to achieve the respective behaviour:
====================

SIMULATION:  If SIMULATION is defined, certain hardware waits and optimizations are turned off.
             SIMULATION must be undefined for final hardware.

DEBUG:       DEBUG will replace inline functions with standard functions. __inline is not supported
             if optimizations are entirely disabled.

XTAL_FRQ:    Crystal frequency in kHz. Supported values:
             XTAL_FRQ==8000
             XTAL_FRQ==11059
             XTAL_FRQ==12000
             XTAL_FRQ==14746
             XTAL_FRQ==18432
             XTAL_FRQ==20000

USE_TEMPSENSOR: Use TMP275 temperature sensor

Note:        Nested interrupts are not permitted!
=====        It is assumed that an interrupt handler is never interrupted.

Fuses:       Set CKSEL[3..0]=0111 for using external crystal oscillator
======       Set SUT[1..0]=11 for long delay from reset, just to be safe
             Tool: "Full Swing Oscillator; Start-up time: 16K CK + 65 ms; Crystal Osc.; slowly rising power"
             Set CKDIV8=0 (programmed) to start with 1/8th of the oscillator frequency
             Tool: CKDIV8 checked (Note: max ISP clock: 250 kHz)
             Prescaler ist setup properly during initialization
             WDTON must not be programmed, watchdog timer is disabled at startup.

ToDo:
=====
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Includes                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#define DECLARE_VARS /* This is the main program in which the global variables are declared */
#include "project.h"
#include "modbus.h"

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
Port Pin connections of
ATMEGA644PA-AU Board with RS485 Interface
=========================================

        | dedicated | I/O | Device | Board | Antenna switch
Port    | Signal    | cfg |  Pin   |  Pin  | used for
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
PA0     |     -     |  I  |    37  |    3  | Measure VIN via 4k7 - 1k voltage divider
PA1     |     -     |  O  |    36  |    4  | Relay for antenna 1
PA2     |     -     |  O  |    35  |    5  | Relay for antenna 2
PA3     |     -     |  O  |    34  |    6  | Relay for antenna 3
PA4     |     -     |  O  |    33  |    7  | Relay for antenna 4
PA5     |     -     |  O  |    32  |    8  | Relay for antenna 5
PA6     |     -     |  O  |    31  |    9  | Relay for antenna 6
PA7     |     -     |  O  |    30  |   10  | unused
PB0     |  DE       |  O  |    40  |    -  | RS485 driver enable
PB1     |     -     |  O  |    41  |   32  | unused
PB2     |     -     |  O  |    42  |   30  | unused
PB3     |  SCL      |  O  |    43  |   29  | Temperature sensor
PB4     |  SDA      |  O* |    44  |    -  | Temperature sensor
PB5     |  MOSI     |  I* |     1  |   27  | unused
PB6     |  MISO     |  I* |     2  |   28  | unused
PB7     |  SCK      |  I* |     3  |   31  | unused
PC0     |     -     |  O  |    19  |   19  | unused, not connected
PC1     |     -     |  O  |    20  |   18  | unused, not connected
PC2     |     -     |  O  |    21  |   17  | unused, not connected
PC3     |     -     |  O  |    22  |   16  | unused, not connected
PC4     |     -     |  O  |    23  |   15  | unused, not connected
PC5     |     -     |  O  |    24  |   14  | unused, not connected
PC6     |     -     |  O  |    25  |   13  | unused, not connected
PC7     |     -     |  O  |    26  |   11  | unused
PD0     |  RXD0     |  I  |     9  |    -  | unused
PD1     |  TXD0     |  O  |    10  |    -  | unused
PD2     |     -     |  O  |    11  |   25  | unused, not connected
PD3     |     -     |  O  |    12  |   24  | unused, not connected
PD4     |     -     |  O  |    13  |   23  | unused, not connected
PD5     |     -     |  O  |    14  |   22  | unused, not connected
PD6     |     -     |  O  |    15  |   21  | unused, not connected
PD7     |     -     |  O  |    16  |   20  | unused, not connected

I*: Input with internal pull-up
O*: Open-drain output

*/
/* Port settings according to previous table */
#define PORTA_DIR    0xfe /* PA7..PA1 are outputs */
#define PORTA_MASK   0x7e /* PA6..PA1 are function outputs */
#define PORTA_PULLUP 0x00 /* no pullups */
#define PORTB_DIR    0x1f /* PB4..PB0 are outputs */
#define PORTB_PULLUP 0xf0 /* PB7..PB4 have pullups */
#define PORTC_DIR    0xff /* PC7..PC0 are outputs */
#define PORTC_PULLUP 0x00 /* no pullups */
#define PORTD_DIR    0xfe /* PD7..PD1 are outputs, PD0 is input */
#define PORTD_PULLUP 0x00 /* no pullups */

/*
+----------------------------+
|       ModBus parameters    |
+----------------------------+
*/
#define USE_TEMPSENSOR
#define USE_ADC
#define ADC_READ_INTERVAL 500    /* ADC read interval in cs */
#define TS_READ_INTERVAL  200    /* Temperature sensor read interval in cs */

#define MODBUS_ADDR 17   /* Default Modbus address of this device */
#define VIRTUAL_ADDRESS_PORTS 1000 /* virtual address of ports for Modbus addressing */
#define VIRTUAL_ADDRESS_IR  2000   /* virtual address of input registers for Modbus addressing (see input_registers in PROGMEM) */
#define VIRTUAL_ADDRESS_HR 3000    /* virtual address of holding (input/output) registers for Modbus addressing (see below) */
#define VA_FR   1000               /* File Records */

#define VA_PORTS  (VIRTUAL_ADDRESS_PORTS +    0) /* Virtual address for I/O port access ("coils") */
#define VA_IR     (VIRTUAL_ADDRESS_IR    +    0) /* Virtual address of input registers */
#define VA_EEP    (VIRTUAL_ADDRESS_IR    +  100) /* Virtual address to read EEPROM contents */
#define VA_FLASH  (VIRTUAL_ADDRESS_IR    + 2000) /* Virtual address to read FLASH memory contents */
#define VA_MCB    (VIRTUAL_ADDRESS_HR    +    0) /* Virtual address of MCB entry (which is a copy from the current EEPROM MCB) */
#define VA_RTC    (VIRTUAL_ADDRESS_HR    +   10) /* Virtual address of first RTC register */
#define VA_SYSREG (VIRTUAL_ADDRESS_HR    +   20) /* Registers for system functions, e.g. restart device */
#define VA_STATUS (VIRTUAL_ADDRESS_HR    +  100) /* Start address for virtual status registers */
#define VA_TELEM  (VIRTUAL_ADDRESS_HR    +  200) /* Start address for telemetric data, temperature, voltage etc. */

/* VA_PORTS + 1: Antenna 1 */
/* VA_PORTS + 2: Antenna 2 */
/* VA_PORTS + 3: Antenna 3 */
/* VA_PORTS + 4: Antenna 4 */
/* VA_PORTS + 5: Antenna 5 */
/* VA_PORTS + 6: Antenna 6 */

#define EEP_SIZE (8u * 128u) /* size of EEPROM in words (2 bytes) */
#define FLASH_SIZE (32u * 1024u) /* size of program memory in words (2 bytes) */
#define MAGIC_NUMBER 0x39f1                         /* magic number to restart device */

#define BOARD_RS485ATM_V10 0x02 /* High byte: board type. RS485 Board with ATMEGA644, Version 1.0 */
#define FUNCTION_ANTSW     0x01 /* Low byte: function. Antenna switch */

#define MODULE_TYPE ((BOARD_RS485ATM_V10<<8)|FUNCTION_ANTSW) /* Module type */
#define HW_VERSION 1000 /* Hardware version (16 bit). Displayed as nn.mmm */
#define SW_VERSION 1000 /* Software version (16 bit). Displayed as nn.mmm */

#define VALID 0x3d      /* Valid marker for MCB (don't use 0x00 or 0xff, any other value should be fine) */

#if (XTAL_FRQ>=400) && (XTAL_FRQ<=900)
#define OSC_OPT ((0<<CKSEL2)|(0<<CKSEL1))       /* 0.4 .. 0.9 MHz */
#endif

#if (XTAL_FRQ>900) && (XTAL_FRQ<=3000)
#define OSC_OPT ((0<<CKSEL2)|(1<<CKSEL1))       /* 0.9 .. 3 MHz */
#endif

#if (XTAL_FRQ>3000) && (XTAL_FRQ<=8000)
#define OSC_OPT ((1<<CKSEL2)|(0<<CKSEL1))       /* 3 .. 8 MHz */
#endif

#if (XTAL_FRQ>8000)
#define OSC_OPT ((1<<CKSEL2)|(1<<CKSEL1))       /* > 8 MHz */
#endif

#define CPU_CLK  XTAL_FRQ /* CPU clock speed in kHz */
#define IO_CLK   XTAL_FRQ /* I/O clock speed in kHz */
#define NS_PER_CPUCLK (1000000l / CPU_CLK) /* nanoseconds per CPU clock cycle */


#ifdef USE_TEMPSENSOR
/*
Timing requirements (of TMP275 in fast mode):
max clock speed: 400 kHz
bus free time between STOP and START: min 1300 ns
Data hold time after SCL low: 4 .. 900 ns
==> ensured by SCL low period
Data setup time between setting SDA and SCL high: 100 ns (min)
==> insert just one NOP delay if clock speed is higher than 10 MHz
Clock low period:  1300 ns (min)
CLOCK high period:  600 ns (min)
SDA must be stable while SCL is high
*/

#define I2C_SCL (1<<PB3)        /* PB3 is I2C SCL signal */
#define I2C_SDA (1<<PB4)        /* PB4 is I2C SDA signal */
#define I2C_SDAPOS PB4          /* PB4 is I2C SDA signal */
#define TS0_ADDR 0x48           /* I2C address of temperature sensor 0 */

uint8_t i2c_read8(uint8_t addr) ;
uint16_t i2c_read16(uint8_t addr) ;
bool i2c_write8(uint8_t addr, uint8_t pointer, uint8_t data) ;
bool i2c_write16(uint8_t addr, uint8_t pointer, uint16_t data) ;
bool i2c_setptr(uint8_t addr, uint8_t pointer) ;

/* I2C interface functions */
__inline static void I2C_PORTCONFIG(void)
                          { /* Configure ports for I2C channel (PB3 for SCL, PB4 for SDA) */
                             PORTB |=  I2C_SCL ; /* SCL is high when idle */
                             PORTB &= ~I2C_SDA ; /* SDA is low when output is enabled */
                             DDRB |=   I2C_SCL ; /* SCL is output */
                             DDRB &=  ~I2C_SDA ; /* SDA is input for now */
                          }

__inline static void I2C_SU_TIME(void)  {if (CPU_CLK > 10000) _NOP() ;} /* ensure setup time */
__inline static void I2C_SCL_LOW(void)  {_delay_loop_1(1+(1300/(3*NS_PER_CPUCLK))) ;} /* ensure min SCL low time */
__inline static void I2C_SCL_HIGH(void) {_delay_loop_1(1+(600/(3*NS_PER_CPUCLK))) ;} /* ensure min SCL high time */
__inline static void I2C_BUSFREE(void)  {_delay_loop_1(1+(1300/(3*NS_PER_CPUCLK))) ;} /* bus free time between stop and start (I2C Spec: 1300ns min) */

__inline static void I2C_SDALOW(void)  {DDRB  |=  I2C_SDA ; I2C_SU_TIME() ;} /* SDA low */
__inline static void I2C_SDAHIGH(void) {DDRB  &= ~I2C_SDA ; I2C_SU_TIME() ;} /* SDA high */
__inline static void I2C_SDASET(uint8_t data) {DDRB = (DDRB & ~I2C_SDA) | ((~data & 0x01)<<I2C_SDAPOS) ; I2C_SU_TIME() ;} /* SDA high */
__inline static void I2C_SCLLOW(void)  {PORTB &= ~I2C_SCL ; _delay_loop_1(1+(1300/(3*NS_PER_CPUCLK))) ;} /* SCL low */
__inline static void I2C_SCLHIGH(void) {PORTB |=  I2C_SCL ; _delay_loop_1(1+(600/(3*NS_PER_CPUCLK))) ;} /* SCL high */

__inline static uint8_t I2C_SDAGET(void) {return ((PINB & I2C_SDA)>>I2C_SDAPOS) ;} /* Read SDA and move to bit 0 */
__inline static void I2C_START(void) {I2C_SDALOW() ; I2C_SCLLOW() ;} /* generate I2C start condition */
__inline static void I2C_STOP(void)  {I2C_SDALOW() ; I2C_SCLHIGH() ; I2C_SDAHIGH() ; I2C_BUSFREE() ;} /* generate I2C stop condition */

#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       structs and enums                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/
typedef struct tim {
        volatile uint16_t timer ;       /* down counter */
        volatile bool active ;          /* status */
        } TIMER ;

typedef struct mcb_struct { /* main control block */
        uint8_t valid ;                 /* This MCB is valid if valid == VALID, any other value means block is invalid.
                                           This must be the FIRST element in the structure! */
        uint8_t module_address ;        /* modbus address of this module */
        uint8_t baudrate ;              /* baudrate, index into "comm_times" table */
        uint8_t porta_state ;           /* state of port A outputs */
        int16_t rtc_correction ;        /* correction for RTC */
        uint16_t pgm_cntrl ;            /* counter for number of EEPROM programming cycles, lower */
        uint16_t pgm_cntrh ;            /* counter for number of EEPROM programming cycles, higher */
        uint16_t crc ;                  /* CRC on all previous elements in MCB, algorithm as for Modbus.
                                           This must be the LAST element in the structure! */
        } MCB ;

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Function prototypes                                                                                             |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void static init_CPU(void) ;
void initial_things(void) ;
void prepare_response(RXBUFFER *rxbuf, TXBUFFER *txbuf) ;
void error_response(TXBUFFER *txbuf, uint8_t errorcode) ;
bool mcb_eepread(MCB *eemcb, MCB *mcb, uint8_t num_blocks) ;
void mcb_pgmread(MCB *mcb, const MCB *pgmmcb) ;
void mcb_eepinit(MCB *eemcb, const MCB *pgmmcb) ;
bool mcb_eepwrite(MCB *eemcb, MCB *mcb, uint8_t num_blocks) ;
extern void wdt_off(void) ;
extern void wdt_on(void) ;
extern void issue_wdt_reset(void) ;
uint16_t restart(uint16_t addr, uint16_t data) ;
uint16_t clear_counters(uint16_t addr, uint16_t data) ;
uint16_t reset_deviceid(uint16_t addr, uint16_t data) ;
uint16_t read_mcb(uint16_t data) ;
uint16_t write_mcb(uint16_t addr, uint16_t data) ;
uint16_t readcoil(uint16_t data) ;
uint16_t writecoil(uint16_t addr, uint16_t data) ;
extern void set_system_clock_prescaler(void) ;
bool i2c_sendbyte(uint8_t data) ;
uint8_t i2c_receivebyte(bool lastbyte) ;

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       EEPROM data                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
*/
MCB EEMEM eemcb[21] ;

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       global variables                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/
MCB mcb ; /* runtime copy of MCB from EEPROM */

/* runtime copies from EEPROM */
uint8_t baudrate ; /* baudrate, index into "comm_times" table */
uint8_t porta_state ; /* state of port A outputs */
bool module_address_updated ; /* true if module_address was updated */
bool baudrate_updated ; /* true if baudrate was updated */
bool porta_state_updated ; /* true if porta_state was updated */

/* RTC is corrected by an extra increment or decrement to the 10ms-counter (rtc_cs) every
   rtc_correction_counter seconds. */
volatile uint16_t rtc_correction_counter ; /* down counter of seconds for RTC correction */
uint16_t rtc_correction_reload ; /* reload value for expired rtc_correction_counter */
int8_t rtc_correction ; /* RTC is corrected forward if positive, backward if negative, no correction if zero */
volatile bool inhibit_rtc_correction ; /* temporarily inhibits RTC correction to prevent double correction */
int16_t temperature ; /* current temperature */
uint16_t vin_voltage ; /* supply voltage */

/* NOINIT: the following variables survive the startup code */
/* Module status and error counters. May be read and written via ModBus functions */
bool software_reset __attribute__ ((section (".noinit"))) ; /* flag to indicate an intentional software reset */
uint8_t reset_flags __attribute__ ((section (".noinit"))) ; /* Copy of reset flags from MCU status register */
uint16_t software_reset_count __attribute__ ((section (".noinit"))) ; /* Number of software resets since last power-on reset */
uint16_t wdt_reset_count __attribute__ ((section (".noinit"))) ; /* Number of watchdog timer resets since last power-on reset */
uint16_t reset_count __attribute__ ((section (".noinit"))) ; /* Number of resets since last power-on reset */
uint16_t ext_reset_count __attribute__ ((section (".noinit"))) ; /* Number of external resets since last power-on reset */
uint16_t bo_reset_count __attribute__ ((section (".noinit"))) ; /* Number of brown-out resets since last power-on reset */

/* RTC counters */
volatile bool rtc_active   __attribute__ ((section (".noinit"))) ; /* RTC is active if true */
volatile bool rtc_midnight __attribute__ ((section (".noinit"))) ; /* Midnight */
volatile uint8_t rtc_cs    __attribute__ ((section (".noinit"))) ; /* cs counter (0..99) */
volatile uint8_t rtc_sec   __attribute__ ((section (".noinit"))) ; /* seconds counter (0..59) */
volatile uint8_t rtc_min   __attribute__ ((section (".noinit"))) ; /* minutes counter (0..59) */
volatile uint8_t rtc_hour  __attribute__ ((section (".noinit"))) ; /* hours counter (0..23) */
uint8_t rtc_day   __attribute__ ((section (".noinit"))) ; /* days counter (1..31) */
uint8_t rtc_month __attribute__ ((section (".noinit"))) ; /* months counter (1..12) */
uint8_t rtc_year  __attribute__ ((section (".noinit"))) ; /* years counter (0..255 = 2000..2255) */

/* Timer */
#ifdef USE_TEMPSENSOR
TIMER ts_timer ;   /* timer for temperature sensor */
TIMER adc_timer ;  /* timer for ADC */
#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       PROGMEM constants                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/
static const PROGMEM MCB default_mcb = {/* Master Control Block */
        .valid = VALID,                 /* this block is valid */
        .module_address = MODBUS_ADDR,  /* modbus address of this module */
        .baudrate = BAUDRATE,           /* baudrate, index into "comm_times" table */
        .porta_state = 0,               /* all switched off */
        .rtc_correction = 0,            /* no RTC correction */
        .pgm_cntrl = 1,                 /* lower programming counter = 1 */
        .pgm_cntrh = 0,                 /* higher programming counter = 0 */
        .crc = 0                        /* unused, will be calculated at runtime */
        } ;

char const PROGMEM FileName[] = __FILE__ ;       /* Filename */
char const PROGMEM TimeStamp[] = __TIMESTAMP__ ; /* Sun Sep 16 01:03:52 1973 */
char const PROGMEM CompileDate[] = __DATE__ ;    /* Mmm dd yyyy */
char const PROGMEM CompileTime[] = __TIME__ ;    /* hh:mm:ss */

                              // ,---------------------------------------- modbus start address
                              // |             ,-------------------------- number of elements
                              // |             |   ,---------------------- type of register
                              // |             |   |  ,------------------- direct value, pointer to value or physical address in case of EEPROM or PROGMEM
                              // |             |   |  |  ,---------------- pointer to read function
const PROGMEM MODBUS_REGISTER // |             |   |  |  |          ,----- pointer to write function
      Coils[] =          {    // |             |   |  |  |          |
                                {VA_PORTS + 0, 24, 0, 0, &readcoil, &writecoil},
                                } ;

                              // ,-------------------------------------------------------------------- modbus start address
                              // |         ,---------------------------------------------------------- number of elements
                              // |         |           ,---------------------------------------------- type of register
                              // |         |           |                     ,------------------------ direct value, pointer to value or physical address in case of EEPROM or PROGMEM
                              // |         |           |                     |            ,----------- pointer to read function
const PROGMEM MODBUS_REGISTER // |         |           |                     |            |     ,----- pointer to write function
      Input_Register[] = {    // |         |           |                     |            |     |
                                {VA_IR+ 0, 1,          T_IMMEDIATE | T_WORD, XTAL_FRQ,    NULL, NULL},  /* crystal frequency in kHz */
                                {VA_IR+ 1, 1,          T_IMMEDIATE | T_WORD, MODULE_TYPE, NULL, NULL},  /* Module type */
                                {VA_IR+ 2, 1,          T_IMMEDIATE | T_WORD, HW_VERSION,  NULL, NULL},  /* Hardware version (16 bit). Displayed as nn.mmm */
                                {VA_IR+ 3, 1,          T_IMMEDIATE | T_WORD, SW_VERSION,  NULL, NULL},  /* Software version (16 bit). Displayed as nn.mmm */
                                {VA_IR+ 4, 1,          T_IMMEDIATE | T_WORD, CPU_CLK,     NULL, NULL},  /* CPU clock speed in kHz */
                                {VA_IR+ 5, 1,          T_IMMEDIATE | T_WORD, IO_CLK,      NULL, NULL},  /* I/O clock speed in kHz */
                                {VA_EEP,   EEP_SIZE,   T_EEPPTR    | T_WORD, 0,           NULL, NULL},
                                {VA_FLASH, FLASH_SIZE, T_PROGPTR   | T_WORD, 0,           NULL, NULL},
                                } ;

                              // ,----------------------------------------------------------------- file record (file number is always 1)
                              // |          ,------------------------------------------------------ number of characters or 0 for null-terminated string
                              // |          |  ,--------------------------------------------------- type of register
                              // |          |  |          ,---------------------------------------- pointer to file record in PROGMEM
                              // |          |  |          |                            ,----------- pointer to read function
const PROGMEM MODBUS_REGISTER // |          |  |          |                            |     ,----- pointer to write function
      File_Records[] = {      // |          |  |          |                            |     |
                                {VA_FR + 0, 0, T_PROGPTR, (uint16_t) FileName,         NULL, NULL},  /* Filename */
                                {VA_FR + 1, 0, T_PROGPTR, (uint16_t) TimeStamp,        NULL, NULL},  /* Sun Sep 16 01:03:52 1973 */
                                {VA_FR + 2, 0, T_PROGPTR, (uint16_t) Modbus_FileName,  NULL, NULL},  /* Modbus Filename */
                                {VA_FR + 3, 0, T_PROGPTR, (uint16_t) Modbus_TimeStamp, NULL, NULL},  /* Sun Sep 16 01:03:52 1973 */
                                {VA_FR + 4, 0, T_PROGPTR, (uint16_t) CompileDate,      NULL, NULL},  /* Mmm dd yyyy */
                                {VA_FR + 5, 0, T_PROGPTR, (uint16_t) CompileTime,      NULL, NULL},  /* hh:mm:ss */
                                } ;


                              // ,--------------------------------------------------------------------------------------------- modbus start address
                              // |             ,------------------------------------------------------------------------------- number of elements
                              // |             |  ,---------------------------------------------------------------------------- type of register
                              // |             |  |                         ,-------------------------------------------------- direct value, pointer to value or physical address in case of EEPROM or PROGMEM
                              // |             |  |                         |                                  ,--------------- pointer to read function
const PROGMEM MODBUS_REGISTER // |             |  |                         |                                  |         ,----- pointer to write function
      Holding_Register[] = {  // |             |  |                         |                                  |         |
                                {VA_MCB+ 0,    1, T_MCBPTR,                 (uint16_t) &mcb.module_address,    read_mcb, write_mcb},       /* Access ModBus address in EEPROM */
                                {VA_MCB+ 1,    1, T_MCBPTR,                 (uint16_t) &mcb.baudrate,          read_mcb, write_mcb},       /* Access baudrate in EEPROM */
                                {VA_MCB+ 2,    1, T_MCBPTR,                 (uint16_t) &mcb.porta_state,       read_mcb, write_mcb},       /* Access Port A status in EEPROM */
                                {VA_MCB+ 3,    1, T_MCBPTR | T_WORD,        (uint16_t) &mcb.rtc_correction,    read_mcb, write_mcb},       /* Access RTC correction in EEPROM */
                                {VA_MCB+ 4,    1, T_MCBPTR | T_RO | T_WORD, (uint16_t) &mcb.pgm_cntrh,         read_mcb, write_mcb},       /* Access programming counter (high) in EEPROM */
                                {VA_MCB+ 5,    1, T_MCBPTR | T_RO | T_WORD, (uint16_t) &mcb.pgm_cntrl,         read_mcb, write_mcb},       /* Access programming counter (low) in EEPROM */
//                               |             |                            |                                  |         |
                                {VA_RTC+ 0,    1, T_DATAPTR,                (uint16_t) &rtc_year,              NULL,     NULL},            /* RTC year */
                                {VA_RTC+ 1,    1, T_DATAPTR,                (uint16_t) &rtc_month,             NULL,     NULL},            /* RTC month */
                                {VA_RTC+ 2,    1, T_DATAPTR,                (uint16_t) &rtc_day,               NULL,     NULL},            /* RTC day */
                                {VA_RTC+ 3,    1, T_DATAPTR,                (uint16_t) &rtc_hour,              NULL,     NULL},            /* RTC hour */
                                {VA_RTC+ 4,    1, T_DATAPTR,                (uint16_t) &rtc_min,               NULL,     NULL},            /* RTC minute */
                                {VA_RTC+ 5,    1, T_DATAPTR,                (uint16_t) &rtc_sec,               NULL,     NULL},            /* RTC seconds */
                                {VA_RTC+ 6,    1, T_DATAPTR,                (uint16_t) &rtc_cs,                NULL,     NULL},            /* RTC centiseconds */
                                {VA_RTC+ 7,    1, T_DATAPTR,                (uint16_t) &rtc_active,            NULL,     NULL},            /* Enable RTC */
//                               |             |                            |                                  |         |
                                {VA_SYSREG+ 0, 1, T_EXECUTE,                0,                                 NULL,     restart},         /* Write MAGIC_NUMBER to restart device */
                                {VA_SYSREG+ 1, 1, T_EXECUTE,                0,                                 NULL,     clear_counters},  /* Write MAGIC_NUMBER to clear error counters */
                                {VA_SYSREG+ 2, 1, T_EXECUTE,                0,                                 NULL,     reset_deviceid},  /* Reset device ID in EEPROM to recover from multiple devices with same ID */
//                               |             |                            |                                  |         |
                                {VA_STATUS+ 0, 1, T_DATAPTR,                (uint16_t) &reset_flags,           NULL,     NULL},            /* Read MCU Status Register */
                                {VA_STATUS+ 1, 1, T_DATAPTR | T_WORD,       (uint16_t) &reset_count,           NULL,     NULL},            /* Read total number of resets since last power-on reset */
                                {VA_STATUS+ 2, 1, T_DATAPTR | T_WORD,       (uint16_t) &wdt_reset_count,       NULL,     NULL},            /* Read number of watchdog timer resets since last power-on reset */
                                {VA_STATUS+ 3, 1, T_DATAPTR | T_WORD,       (uint16_t) &software_reset_count,  NULL,     NULL},            /* Read number of software resets since last power-on reset */
                                {VA_STATUS+ 4, 1, T_DATAPTR | T_WORD,       (uint16_t) &ext_reset_count,       NULL,     NULL},            /* Read number of external resets since last power-on reset */
                                {VA_STATUS+ 5, 1, T_DATAPTR | T_WORD,       (uint16_t) &bo_reset_count,        NULL,     NULL},            /* Read number of brown-out resets since last power-on reset */
                                {VA_STATUS+ 6, 1, T_DATAPTR | T_WORD,       (uint16_t) &total_message_count_h, NULL,     NULL},            /* Read total number of error free messages since last power-on reset (higher) */
                                {VA_STATUS+ 7, 1, T_DATAPTR | T_WORD,       (uint16_t) &total_message_count_l, NULL,     NULL},            /* Read total number of error free messages since last power-on reset (lower) */
                                {VA_STATUS+ 8, 1, T_DATAPTR | T_WORD,       (uint16_t) &message_count_h,       NULL,     NULL},            /* Read number of messages for this module since last power-on reset (higher) */
                                {VA_STATUS+ 9, 1, T_DATAPTR | T_WORD,       (uint16_t) &message_count_l,       NULL,     NULL},            /* Read number of messages for this module since last power-on reset (lower) */
                                {VA_STATUS+10, 1, T_DATAPTR | T_WORD,       (uint16_t) &crc_error_count_h,     NULL,     NULL},            /* Read number of CRC errors since last power-on reset (higher) */
                                {VA_STATUS+11, 1, T_DATAPTR | T_WORD,       (uint16_t) &crc_error_count_l,     NULL,     NULL},            /* Read number of CRC errors since last power-on reset (lower) */
                                {VA_STATUS+12, 1, T_DATAPTR | T_WORD,       (uint16_t) &com_error_count_h,     NULL,     NULL},            /* Read number of communication errors since last power-on reset (higher) */
                                {VA_STATUS+13, 1, T_DATAPTR | T_WORD,       (uint16_t) &com_error_count_l,     NULL,     NULL},            /* Read number of communication errors since last power-on reset (lower) */
                                {VA_STATUS+14, 1, T_DATAPTR | T_WORD,       (uint16_t) &com_rx_count_h,        NULL,     NULL},            /* Read number of received bytes since last power-on reset (higher) */
                                {VA_STATUS+15, 1, T_DATAPTR | T_WORD,       (uint16_t) &com_rx_count_l,        NULL,     NULL},            /* Read number of received bytes since last power-on reset (lower) */
                                {VA_STATUS+16, 1, T_DATAPTR | T_WORD,       (uint16_t) &com_tx_count_h,        NULL,     NULL},            /* Read number of transmitted bytes since last power-on reset (higher) */
                                {VA_STATUS+17, 1, T_DATAPTR | T_WORD,       (uint16_t) &com_tx_count_l,        NULL,     NULL},            /* Read number of transmitted bytes since last power-on reset (lower) */
//                               |             |                            |                                  |         |
                                {VA_TELEM+  0, 1, T_DATAPTR | T_RO | T_WORD,(uint16_t) &temperature,           NULL,     NULL},            /* Read temperature */
                                {VA_TELEM+  1, 1, T_DATAPTR | T_RO | T_WORD,(uint16_t) &vin_voltage,           NULL,     NULL}             /* Read supply voltage Vin */
        } ;

const PROGMEM MODBUS_DESCRIPTOR modbus_desc = {
        (PROGMEM MODBUS_REGISTER *) Input_Register,   sizeof(Input_Register)/sizeof(Input_Register[0]),
        (PROGMEM MODBUS_REGISTER *) Holding_Register, sizeof(Holding_Register)/sizeof(Holding_Register[0]),
        (PROGMEM MODBUS_REGISTER *) Coils,            sizeof(Coils)/sizeof(Coils[0]),
        (PROGMEM MODBUS_REGISTER *) File_Records,     sizeof(File_Records)/sizeof(File_Records[0])
        } ;

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Inline functions                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/
_INLINE_ void eeprom_write_byte_cond(uint8_t  *addr, uint8_t  data) {if (eeprom_read_byte(addr) != data) eeprom_write_byte(addr, data) ; }
_INLINE_ void eeprom_write_word_cond(uint16_t *addr, uint16_t data) {if (eeprom_read_word(addr) != data) eeprom_write_word(addr, data) ; }

#ifdef SIMULATION
_INLINE_ static void go_idle(void) {} /* do nothing */
#else
__inline static void go_idle(void) {SMCR = (0<<SM2) | (0<<SM1) | (0<<SM0) | (1<<SE) ;    /* select idle mode and enable sleep mode */
                                   __asm("sleep") ;     /* goto sleep mode */
                                   SMCR = 0 ; }         /* clear sleep mode enable */
#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Main                                                                                                            |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| MAIN program.                                                                                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
int main(void)
{ /* executed on coldstart */
bool rc ;

   initial_things() ;

/* read master control block from EEPROM */
   rc = mcb_eepread(&eemcb[0], &mcb, sizeof(eemcb)/sizeof(eemcb[0])) ; /* read mcb from EEPROM to RAM */
   if (!rc)
   { /* No valid mcb found. Initialise EEPROM with default values */
      mcb_eepinit(&eemcb[0], &default_mcb) ;
      mcb_pgmread(&mcb, &default_mcb) ; /* get default mcb from program memory just in case EEPROM is fatally ill */
   }

   module_address = mcb.module_address ;
   baudrate = mcb.baudrate ;
   porta_state = mcb.porta_state ;

   if (mcb.rtc_correction == 0) rtc_correction = 0 ;
   else
   {
      rtc_correction_reload = abs(mcb.rtc_correction) ;
      rtc_correction_counter = abs(mcb.rtc_correction) ;
      rtc_correction = (mcb.rtc_correction > 0) ? 1 : -1 ;
   }

/* get communication timings */
   ubrrval = pgm_read_word(&comm_times[baudrate].ubrr) ;
   t25 = pgm_read_word(&comm_times[baudrate].t25) ;
   t35 = pgm_read_word(&comm_times[baudrate].t35) ;

   init_CPU() ; /* initialize CPU, turn global interrupts on */

   wdt_on() ; /* switch watchdog timer on */

   listen_only = false ;
   restart_communications = false ;

   for (;;)
   {
/* Setup serial receiver and wait for message */
      receive(&rxbuf) ;
      wait_for_rx(&rxbuf) ; /* wait for received message */

      switch(rxbuf.status)
      { /* reception completed. Check for error, otherwise check address and prepare response */
         case RXSTATE_COMPLETED:
            module_address_updated = false ;
            baudrate_updated = false ;
            porta_state_updated = false ;

            modbus_message_received(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;

            if (module_address_updated || baudrate_updated || porta_state_updated)
            { /* MCB in EEPROM must be renewed */
               /* read current master control block from EEPROM */
               rc = mcb_eepread(&eemcb[0], &mcb, sizeof(eemcb)/sizeof(eemcb[0])) ; /* read mcb from EEPROM to RAM */
               if (rc)
               { /* update block only on success, otherwise do nothing */
                  if (module_address_updated) mcb.module_address = module_address ;
                  if (baudrate_updated) mcb.baudrate = baudrate ;
                  if (porta_state_updated) mcb.porta_state = porta_state ;
                  mcb_eepwrite(&eemcb[0], &mcb, sizeof(eemcb)/sizeof(eemcb[0])) ; /* write mcb from RAM to EEPROM, ignore any possible error */
               }
            }
         break ;

         case RXSTATE_ERROR:
         case RXSTATE_BUFFEROVERFLOW :
         /* error, just increment error counter */
            com_error_count_l++ ;
            if (com_error_count_l == 0) com_error_count_h++ ;
         break ;

         default: /* ignore other cases for now */
         break ;
      }
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       initial_things                                                                                                  |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Do initialization after a reset occured.                                                                              |
|                                                                                                                       |
| This function is executed immediatelly after main() is started. The CPU is not yet initialized, it runs at its        |
| default speed (1 MHz). Determine cause of reset and optionally initialize reset- and error counters.                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void initial_things(void)
{
   reset_flags = MCUSR ; /* save Reset flags */
   wdt_off() ; /* Switch off the watchdog timer. WDT may still be enabled from an intentionally issued WDT reset. */
   if ((reset_flags & (1<<PORF)) != 0)
   { /* This is a cold start */
#ifndef SIMULATION
      _delay_loop_2 (25000) ;    /* wait 100 ms for power settling (four CPU cycles per iteration) */
#endif
      software_reset = false ;
      software_reset_count = 0 ;
      ext_reset_count = 0 ;
      bo_reset_count = 0 ;
      wdt_reset_count = 0 ;
      reset_count = 0 ;
      crc_error_count_l = 0 ;
      crc_error_count_h = 0 ;
      message_count_l = 0 ;
      message_count_h = 0 ;
      total_message_count_l = 0 ;
      total_message_count_h = 0 ;
      com_error_count_l = 0 ;
      com_error_count_h = 0 ;
      com_rx_count_l = 0 ;
      com_rx_count_h = 0 ;
      com_tx_count_l = 0 ;
      com_tx_count_h = 0 ;

      rtc_active   = false ;
      rtc_midnight = false ;
      rtc_cs    = 0 ;
      rtc_sec   = 0 ;
      rtc_min   = 0 ;
      rtc_hour  = 0 ;
      rtc_day   = 0 ;
      rtc_month = 0 ;
      rtc_year  = 0 ;
   }

   reset_count++ ;
   if ((reset_flags & (1<<WDRF)) != 0)
   { /* This is a WDT reset */
      if (software_reset)
      { /* this is an intentional software reset */
         software_reset_count++ ;
         software_reset = false ;
      }
      else wdt_reset_count++ ;
   }
   if ((reset_flags & (1<<BORF)) != 0)  bo_reset_count++ ;       /* This is a brown-out reset */
   if ((reset_flags & (1<<EXTRF)) != 0) ext_reset_count++ ;      /* This is a external reset */

   MCUSR = 0 ; /* Clear reset flags in MCU status register */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       wait_for_rx                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Wait for reception of a Modbus command. Do some house keeping while waiting.                                          |
|                                                                                                                       |
| This function executes regularly while waiting for a bus transaction. Interrupts shall not be disabled.               |
| Bus messages may be lost, if execution takes too long.                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void wait_for_rx(RXBUFFER *rxb)
{
uint8_t num_days ;
uint16_t tmp ;

   while ((rxb->status == RXSTATE_RECEIVING) || (rxb->status == RXSTATE_TIMEOUT))
   {
      __asm("wdr") ; /* say hello to the watchdog */

#ifdef USE_TEMPSENSOR
      if (!ts_timer.active)
      { /* read temperature every now and then */
         i2c_setptr(TS0_ADDR, 0) ;
         temperature = ((int16_t) (i2c_read16(TS0_ADDR) & 0xfff0)) / 16 ;
         ts_timer.timer = TS_READ_INTERVAL ;
         ts_timer.active = true ;
      }
#endif

#ifdef USE_ADC
      if (!adc_timer.active)
      { /* read VIN voltage via ADC every now and then */
         if ((ADCSRA & (1<<ADSC)) == 0) /* no conversion ongoing */
            vin_voltage = ADC ; /* read last conversion result, first one will be wrong  */
         /* start next conversion */
         ADCSRA |= (1<<ADSC) ;
         adc_timer.timer = ADC_READ_INTERVAL ;
         adc_timer.active = true ;
      }
#endif

/* do some house keeping */
      if (rtc_midnight)
      { /* update date */
         rtc_midnight = false ;
         /* days per month: 28: Feb; 29: Feb (leap year); 30: Apr, Jun, Sept, Nov; 31: Jan, Mar, May, Jul, Aug, Oct, Dec */
         if (rtc_month == 2)
         { /* it is February, check for leap year. Note that year offset is 2000 */
            num_days = 28 ;
            if (((rtc_year % 4) == 0) && (rtc_year != 100) && (rtc_year != 200)) num_days = 29 ; /* this is a leap year */
         }
         else
         {
            if ((rtc_month == 4) || (rtc_month == 6) || (rtc_month == 9) || (rtc_month == 11)) num_days = 30 ; /* 30 days per month */
            else num_days = 31 ; /* this month has 31 days */
         }
         if (rtc_day < num_days) rtc_day++ ;
         else
         { /* new month */
            rtc_day = 1 ;
            if (rtc_month < 12) rtc_month++ ;
            else
            { /* new year */
               rtc_month = 1 ;
               rtc_year++ ;
            }
         }
      }
      go_idle() ; /* idle until next interrupt, i.e. max. 10 ms. No need to switch off watchdog */
   }

/* data received. Add number of received bytes to total count */
   tmp = com_rx_count_l + rxb->bytecounter ;
   if (tmp < com_rx_count_l) com_rx_count_h++ ; /* we had an overflow in the lower half */
   com_rx_count_l = tmp ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       wait_for_tx                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Wait for transmit completion                                                                                          |
|                                                                                                                       |
| This function executes regularly while waiting for a bus transaction. Interrupts shall not be disabled.               |
| Bus messages may be lost, if execution takes too long.                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void wait_for_tx(TXBUFFER *txb)
{
uint16_t tmp ;
   while (!txb->completed)
   { /* wait for completion */
      __asm("wdr") ; /* say hello to the watchdog */
      _delay_loop_1 (10) ; /* wait a little while (three CPU cycles per iteration) */
   }

/* all data sent. Add number of sent bytes to total count */
   tmp = com_tx_count_l + txb->bytecounter ;
   if (tmp < com_tx_count_l) com_tx_count_h++ ; /* we had an overflow in the lower half */
   com_tx_count_l = tmp ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       mcb_eepread                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Read currently active mcb from EEPROM into RAM                                                                        |
| "eemcb" points to the first mcb in the EEPROM (source address).                                                       |
| "mcb" points to the mcb in RAM (target address).                                                                      |
| "num_blocks" is the size of the mcb array in EEPROM.                                                                  |
|                                                                                                                       |
| Return true if valid block was read, otherwise return false.                                                          |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool mcb_eepread(MCB *eemcb, MCB *mcb, uint8_t num_blocks)
{
uint8_t i, j ;
uint8_t val ;
uint8_t *src ;
uint8_t *target ;
uint16_t crc ;
bool rc ;
   rc = false ;
   for (i=0; i<num_blocks; i++)
   {
      val = eeprom_read_byte(&eemcb[i].valid) ;
      if (val == VALID)
      { /* valid entry found, retrieve data */
         src = (uint8_t *) &eemcb[i] ;
         target = (uint8_t *) mcb ;
         for (j=0; j<sizeof(MCB); j++) target[j] = eeprom_read_byte(&src[j]) ;
         crc = modbus_crc16(target, sizeof(MCB)-2) ; /* generate CRC */
         rc = (crc == mcb->crc) ;
      }
   }
   return(rc) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       mcb_eepwrite                                                                                                    |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Write mcb from RAM into new mcb in EEPROM. Increment number of write cycles.                                          |
| "eemcb" points to the first mcb in the EEPROM (target address).                                                       |
| "mcb" points to the mcb in RAM (source address).                                                                      |
| "num_blocks" is the size of the mcb array in EEPROM.                                                                  |
|                                                                                                                       |
| Return true if block was successfully written, otherwise return false.                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool mcb_eepwrite(MCB *eemcb, MCB *mcb, uint8_t num_blocks)
{
uint8_t i ;
uint8_t val ;
uint8_t *src ;
uint8_t *oldmcb ;
uint8_t *newmcb ;
uint16_t crc ;
uint16_t countl ;
uint16_t counth ;
bool rc ;
   rc = false ;
   for (i=0; ((i<num_blocks) && !rc); i++)
   { /* find first valid block */
      val = eeprom_read_byte(&eemcb[i].valid) ;
      if (val == VALID)
      { /* valid block found */
         rc = true ;
         oldmcb = (uint8_t *) &eemcb[i] ; /* save its address */
         countl = eeprom_read_word(&eemcb[i].pgm_cntrl) ;
         counth = eeprom_read_word(&eemcb[i].pgm_cntrh) ;
         if (i < (num_blocks-1))
            newmcb = (uint8_t *) &eemcb[i+1] ; /* pointer to next mcb */
         else
            newmcb = (uint8_t *) &eemcb[0] ; /* wrap over to first mcb */
      }
   }
   if (rc)
   { /* old and new address found. Increment number of write cycles, generate CRC and write data */
      mcb->valid = VALID ; /* just in case... */
      countl++ ;
      if (countl == 0) counth++ ;
      mcb->pgm_cntrl = countl ;
      mcb->pgm_cntrh = counth ;
      src = (uint8_t *) mcb ;
      crc = modbus_crc16(src, sizeof(MCB)-2) ; /* generate CRC */
      mcb->crc = crc ;
      for (i=1; i<sizeof(MCB); i++) eeprom_write_byte(&newmcb[i], src[i]) ;
      eeprom_write_byte(&newmcb[0], src[0]) ; /* stamp the new mcb valid */
      eeprom_write_byte(&oldmcb[0], ~VALID) ; /* finally stamp the old mcb invalid */
   }
   return(rc) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       mcb_eepinit                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Initialise mcb in EEPROM with default values from program memory.                                                     |
| "eemcb" points to the first mcb in the EEPROM. Always only the first mcb will be initialised.                         |
| "pgmmcb" points to the mcb in program memory, which holds the default values.                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void mcb_eepinit(MCB *eemcb, const MCB *pgmmcb)
{
static volatile MCB tmpmcb ;
uint8_t *src ;
uint8_t *target ;
uint16_t crc ;
uint8_t i ;
   src = (uint8_t *) pgmmcb ;
   target = (uint8_t *) &tmpmcb ;
   for (i=0; i<sizeof(MCB); i++) target[i] = pgm_read_byte(&src[i]) ; /* get data from progmem to local buffer */
   crc = modbus_crc16(target, sizeof(MCB)-2) ; /* generate CRC */
   tmpmcb.crc = crc ;
   src = (uint8_t *) &tmpmcb ;
   target = (uint8_t *) &eemcb[0] ; /* target address is block 0 */
   for (i=1; i<sizeof(MCB); i++) eeprom_write_byte(&target[i], src[i]) ; /* write block into EEPROM */
   eeprom_write_byte(&target[0], src[0]) ; /* finally stamp the mcb valid */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       mcb_pgmread                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Read default mcb from program memory to ram.                                                                          |
| "mcb" points to the mcb in RAM (target address).                                                                      |
| "pgmmcb" points to the mcb in program memory, which holds the default values.                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void mcb_pgmread(MCB *mcb, const MCB *pgmmcb)
{
uint8_t *src ;
uint8_t *target ;
uint16_t crc ;
uint8_t i ;
   src = (uint8_t *) pgmmcb ;
   target = (uint8_t *) mcb ;
   for (i=0; i<sizeof(MCB); i++) target[i] = pgm_read_byte(&src[i]) ; /* get data from progmem to ram buffer */
   crc = modbus_crc16(target, sizeof(MCB)-2) ; /* generate CRC */
   mcb->crc = crc ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       read_mcb                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| This is a callback function for the modbus library. It reads the EEPROM MCB into the RAM mcb.                         |
|                                                                                                                       |
| Return true if valid block was read, otherwise return false.                                                          |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint16_t read_mcb(uint16_t data)
{ /* read mcb from EEPROM to RAM */
   return (mcb_eepread(&eemcb[0], &mcb, sizeof(eemcb)/sizeof(eemcb[0]))) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       write_mcb                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| This is a callback function for the modbus library. It writes the RAM mcb back to the EEPROM mcb.                     |
|                                                                                                                       |
| Return true if block was successfully written, otherwise return false.                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint16_t write_mcb(uint16_t addr, uint16_t data)
{ /* write mcb from RAM to EEPROM */
   return (mcb_eepwrite(&eemcb[0], &mcb, sizeof(eemcb)/sizeof(eemcb[0]))) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       init_CPU                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Initialize device                                                                                                     |
|                                                                                                                       |
| This function is executed once after a hardware reset. Do all initialization, which survives a power down mode.       |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void init_CPU(void)
{
#ifndef SIMULATION
/* set clock prescaler */
   set_system_clock_prescaler() ;
#endif

/* Switch off unused peripherals for power reduction */
#ifdef USE_ADC
   PRR0 = (1<<PRTWI) | (1<<PRTIM2) | (0<<PRTIM1) | (1<<PRTIM0) | (1<<PRUSART1) | (0<<PRUSART0) | (1<<PRSPI) | (0<<PRADC) ; /* TIMER1, USART0 and ADC remain on */
#else
   ADCSRA = (0<<ADEN) ; /* disable ADC */
   PRR0 = (1<<PRTWI) | (1<<PRTIM2) | (0<<PRTIM1) | (1<<PRTIM0) | (1<<PRUSART1) | (0<<PRUSART0) | (1<<PRSPI) | (1<<PRADC) ; /* TIMER1 and USART0 remain on */
#endif

/* Initialize ports */
/* Port A */
   PORTA = PORTA_PULLUP|(porta_state & PORTA_MASK) ; /* pull-ups as required, set functional outputs to previous state, other outputs off */
   DDRA  = PORTA_DIR ;          /* enable outputs */

/* Port B */
   PORTB = PORTB_PULLUP|0x00 ;  /* pull-ups as required, all outputs low */
   DDRB  = PORTB_DIR ;          /* enable outputs */

/* Port C */
   PORTC = PORTC_PULLUP|0x00 ;  /* pull-ups as required, all outputs low */
   DDRC  = PORTC_DIR ;          /* enable outputs */

/* Port D */
   PORTD = PORTD_PULLUP|0x00 ;  /* pull-ups as required, all outputs low */
   DDRD  = PORTD_DIR ;          /* enable outputs */

   init_UART() ;

#ifdef USE_TEMPSENSOR
   I2C_PORTCONFIG() ; /* configure I2C interface */
   i2c_write8(TS0_ADDR,1,0x60) ; /* write config register: 12 bits resolution, continuous conversion  */
#endif

#ifdef USE_ADC
   /* use AREF for voltage reference, right adjust, single ended ADC0 */
   ADMUX = (0<<REFS1) | (0<<REFS0) | (0<<ADLAR) | (0<<MUX4) | (0<<MUX3) | (0<<MUX2) | (0<<MUX1) | (0<<MUX0) ;

   /* enable ADC, prescaler = 128 (clock freq must be between 50 and 200 kHz, so this setting is good for 6.4~20 MHz CPU clock */
   ADCSRA = (1<<ADEN) | (0<<ADSC) | (0<<ADATE) | (0<<ADIF) | (0<<ADIE) | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0) ;

   /* use Free running mode */
   ADCSRB = (0<<ADTS2) | (0<<ADTS1) | (0<<ADTS0) ;

   /* Disable digital PA0 input */
   DIDR0 = (0<<ADC7D) | (0<<ADC6D) | (0<<ADC5D) | (0<<ADC4D) | (0<<ADC3D) | (0<<ADC2D) | (0<<ADC1D) | (1<<ADC0D) ;
#endif

/* Timer 1 is used for 10 ms interrupts */
/* Init 16-bit Timer 1 for 10 ms interrupts */
   TCCR1A = 0x00 ;     /* Normal port operation: OC1A/B disconnected, normal waveform generation mode */
   TCCR1B = 0x03 ;     /* normal waveform generation mode, clk(I/O)/64 */
   TCCR1C = 0x00 ;     /* unused */
   TCNT1 = 0 ;         /* make sure to start at 0 */
   OCR1A = RLD_OCR1A  ; /* load compare register for next interrupt in 10 ms */
   TIFR1 = ((1<<ICF1) | (1<<OCF1B) | (1<<OCF1A) | (1<<TOV1)) ; /* clear any possibly pending interrupt */
   TIMSK1 = (1<<OCIE1A) ; /* enable Output Compare 1A interrupt */

   __asm("sei") ; /*  Turn on global interrupt */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       readcoil                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| This is a callback function for the modbus library. It reads one single coil.                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint16_t readcoil(uint16_t addr)
{
/*
Port mapping is as follows:
VA_PORTS+ 0: PA0
VA_PORTS+ 1: PA1 - Antenna 1
VA_PORTS+ 2: PA2 - Antenna 2
VA_PORTS+ 3: PA3 - Antenna 3
VA_PORTS+ 4: PA4 - Antenna 4
VA_PORTS+ 5: PA5 - Antenna 5
VA_PORTS+ 6: PA6 - Antenna 6
VA_PORTS+ 7: PA7
VA_PORTS+ 8: PB0
VA_PORTS+ 9: PB1
VA_PORTS+10: PB2
VA_PORTS+11: PB3
VA_PORTS+12: unused, returns 0
VA_PORTS+13: unused, returns 0
VA_PORTS+14: unused, returns 0
VA_PORTS+15: unused, returns 0
VA_PORTS+16: PC0
VA_PORTS+17: PC1
VA_PORTS+18: PC2
VA_PORTS+19: unused, returns 0
VA_PORTS+20: unused, returns 0
VA_PORTS+21: unused, returns 0
VA_PORTS+22: unused, returns 0
VA_PORTS+23: unused, returns 0
*/
   if ((addr >= (VA_PORTS+0)) && (addr < (VA_PORTS+8)))
      return ((PINA >> (addr - (VA_PORTS+0))) & 1) ; /* Port A access */
   if ((addr >= (VA_PORTS+8)) && (addr < (VA_PORTS+16)))
      return ((PINB >> (addr - (VA_PORTS+8))) & 1) ; /* Port B access */
   if ((addr >= (VA_PORTS+16)) && (addr < (VA_PORTS+24)))
      return ((PINC >> (addr - (VA_PORTS+16))) & 1) ; /* Port C access */
   return(0) ; /* should not happen */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       writecoil                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| This is a callback function for the modbus library. It writes one single coil.                                        |
|                                                                                                                       |
| Returns 0x0000 if data is 0, 0ff00 if data is 1.                                                                      |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint16_t writecoil(uint16_t addr, uint16_t data)
{
uint8_t mask ;

   if ((addr >= (VA_PORTS + 0)) && (addr < (VA_PORTS + 8)))
   { /* Port A */
      mask = (1<<(addr-VA_PORTS)) & PORTA_MASK ;
      if (data != 0) PORTA  = (PORTA & ~PORTA_MASK) | mask ; /* set selected functional port, clear others.
                                                                Just one port can be on. Retain values outside PORTA_MASK */
      else           PORTA &= ~mask ; /* clear port, all others are not affected */
      _delay_loop_1 (10) ;            /* small delay to allow port to switch */
      porta_state = PINA & mask ;
      porta_state_updated = true ;
   }

   return((porta_state==0)?0x0000:0xff00) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       restart                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Code to restart device                                                                                                |
| Returns true if write data does not match the magic number. Will not return otherwise.                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint16_t restart(uint16_t addr, uint16_t data)
{
   if (data == MAGIC_NUMBER)
   {
      software_reset = true ;
      issue_wdt_reset() ; /* Restart module by generating a WDT reset */
   }
   return(true) ; /* should not happen */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       clear_counters                                                                                                  |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Clear various counters                                                                                                |
|                                                                                                                       |
| Returns true if an error occured, false otherwise.                                                                    |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint16_t clear_counters(uint16_t addr, uint16_t data)
{
   if (data == MAGIC_NUMBER)
   {
      software_reset_count = 0 ;
      ext_reset_count = 0 ;
      bo_reset_count = 0 ;
      wdt_reset_count = 0 ;
      reset_count = 0 ;
      total_message_count_l = 0 ;
      total_message_count_h = 0 ;
      message_count_l = 0 ;
      message_count_h = 0 ;
      crc_error_count_l = 0 ;
      crc_error_count_h = 0 ;
      com_error_count_l = 0 ;
      com_error_count_h = 0 ;
      com_rx_count_l = 0 ;
      com_rx_count_h = 0 ;
      com_tx_count_l = 0 ;
      com_tx_count_h = 0 ;
      return(false) ;
   }
   else return(true) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       reset_deviceid                                                                                                  |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Code to reset the device ID                                                                                           |
|                                                                                                                       |
| Returns true if an error occured, false otherwise.                                                                    |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint16_t reset_deviceid(uint16_t addr, uint16_t data)
{
   if (data == MAGIC_NUMBER)
   { /* MCB is about to be updated, read current MCB from EEPROM */
      if (mcb_eepread(&eemcb[0], &mcb, sizeof(eemcb)/sizeof(eemcb[0]))) /* read mcb from EEPROM to RAM */
      { /* Valid mcb found. */
         mcb.module_address = MODBUS_ADDR ;
         mcb_eepwrite(&eemcb[0], &mcb, sizeof(eemcb)/sizeof(eemcb[0])) ; /* write mcb from RAM to EEPROM, ignore any possible error */
      }
      return(false) ;
   }
   else return(true) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       TIMER1_COMPA                                                                                                    |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Interrupt handler for timer1 compare match A. 10 ms interrupt.                                                        |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
ISR(TIMER1_COMPA_vect)
{
   OCR1A += RLD_OCR1A ; /* prepare for next interrupt in 10 ms from this interrupt */

#ifdef USE_TEMPSENSOR
   if (ts_timer.active)
   {
      if (ts_timer.timer != 0) ts_timer.timer-- ;
      if (ts_timer.timer == 0) ts_timer.active = false ;
   }
#endif

#ifdef USE_ADC
   if (adc_timer.active)
   {
      if (adc_timer.timer != 0) adc_timer.timer-- ;
      if (adc_timer.timer == 0) adc_timer.active = false ;
   }
#endif

/* Real Time Clock */
   if (rtc_active)
   {
      if (rtc_correction != 0)
      {
         if (rtc_cs == 60) inhibit_rtc_correction = false ;
         if ((!inhibit_rtc_correction) && (rtc_cs == 50))
         {
            rtc_correction_counter-- ;
            if (rtc_correction_counter == 0)
            { /* apply correction */
               rtc_correction_counter = rtc_correction_reload ;
               if (rtc_correction > 0) rtc_cs++ ;
               if (rtc_correction < 0)
               {
                  rtc_cs-- ;
                  inhibit_rtc_correction = true ; /* prevents double correction if rtc_cs was decremented previously */
               }
            }
         }
      }

      if (rtc_cs == 99)
      { /* next second */
         rtc_cs = 0 ;
         if (rtc_sec == 59)
         {
            rtc_sec = 0 ;
            if (rtc_min == 59)
            {
               rtc_min = 0 ;
               if (rtc_hour == 23)
               {
                  rtc_hour = 0 ;
                  rtc_midnight = true ;
               }
               else rtc_hour++ ;
            }
            else rtc_min++ ;
         }
         else rtc_sec++ ;
      }
      else rtc_cs++ ;
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       TIMER1_COMPB                                                                                                    |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Interrupt handler for communication timeouts.                                                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
ISR(TIMER1_COMPB_vect)
{
   switch(prxb->status)
   {
      case RXSTATE_RECEIVING:
         prxb->status = RXSTATE_TIMEOUT ;
         OCR1B = TCNT1+(t35-t25) ;  /* set frame timeout, t25 is already passed */
      break ;

      case RXSTATE_TIMEOUT:
         prxb->status = RXSTATE_COMPLETED ; /* we had a character timeout before, so this is the end of the frame */
         TOINT_disable() ; /* disable further timeout interrupts */
      break ;

      default:
         TOINT_disable() ; /* disable further timeout interrupts */
      break ;
   }
}

#ifdef USE_TEMPSENSOR

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       i2c_read8                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Read data from an I2C cannel                                                                                          |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint8_t i2c_read8(uint8_t addr)
{
uint8_t data ;
bool ack ;

   data = 0 ;
   I2C_START() ;
   ack = i2c_sendbyte(addr<<1 | 0x01) ; /* transfer address and R/W = high (for read) */
   if (ack) data  = i2c_receivebyte(true) ; /* read data byte */
   I2C_STOP() ;

   return(data) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       i2c_read16                                                                                                      |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Read data from an I2C channel                                                                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint16_t i2c_read16(uint8_t addr)
{
uint16_t data ;
bool ack ;

   data = 0 ;
   I2C_START() ;
   ack = i2c_sendbyte(addr<<1 | 0x01) ; /* transfer address and R/W = high (for read) */
   if (ack)
   {
      data  = ((uint16_t) i2c_receivebyte(false)) << 8 ; /* read MSB */
      data |=  (uint16_t) i2c_receivebyte(true) ; /* read LSB */
   }
   I2C_STOP() ;

   return(data) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       i2c_write8                                                                                                      |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Write 8-bit data to an I2C channel                                                                                    |
|                                                                                                                       |
| Returns true if successful (all ACK properly received)                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool i2c_write8(uint8_t addr, uint8_t pointer, uint8_t data)
{
uint8_t ack ;

   I2C_START() ;
   ack  = i2c_sendbyte(addr<<1) ; /* transfer address and R/W = low (for write) */
   if (ack) ack |= i2c_sendbyte(pointer) ; /* transfer register pointer */
   if (ack) ack |= i2c_sendbyte(data) ; /* transfer byte */
   I2C_STOP() ;

   return(ack == 0) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       i2c_write16                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Write 16-bit data to an I2C channel                                                                                   |
|                                                                                                                       |
| Returns true if successful (all ACK properly received)                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool i2c_write16(uint8_t addr, uint8_t pointer, uint16_t data)
{
uint8_t ack ;

   I2C_START() ;
   ack  = i2c_sendbyte(addr<<1) ; /* transfer address and R/W = low (for write) */
   if (ack) ack |= i2c_sendbyte(pointer) ; /* transfer register pointer */
   if (ack) ack |= i2c_sendbyte(data>>8) ; /* transfer MSB */
   if (ack) ack |= i2c_sendbyte(data&0xff) ; /* transfer LSB */
   I2C_STOP() ;

   return(ack == 0) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       i2c_setptr                                                                                                      |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Set register pointer for I2C                                                                                          |
|                                                                                                                       |
| Returns true if successful (all ACK properly received)                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool i2c_setptr(uint8_t addr, uint8_t pointer)
{
uint8_t ack ;

   I2C_START() ;

   ack  = i2c_sendbyte(addr<<1) ; /* transfer address and R/W = low (for write) */
   if (ack) ack |= i2c_sendbyte(pointer) ; /* transfer pointer */

   I2C_STOP() ;

   return(ack == 0) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       i2c_sendbyte                                                                                                    |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Write byte data and read acknowledge                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool i2c_sendbyte(uint8_t data)
{
uint8_t ack ;
int i ;
   for (i=7; i>=0; i--)
   { /* send bits 7..0 (high order bit first) */
      I2C_SDASET(data>>i) ; /* set SDA */
      I2C_SCLHIGH() ;
      I2C_SCLLOW() ;
   }
   I2C_SDAHIGH() ; /* release pull down to read ack */
   I2C_SCLHIGH() ;
   ack = I2C_SDAGET() ; /* read acknowledge from slave */
//   I2C_SDALOW() ; /* pull SDA down again */
   I2C_SCLLOW() ;

   return(ack == 0) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       i2c_receivebyte                                                                                                 |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Receive byte data and read acknowledge                                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint8_t i2c_receivebyte(bool lastbyte)
{
int i ;
uint8_t data ;
   data = 0 ;
   I2C_SDAHIGH() ; /* release pull down */
   for (i=0; i<8; i++)
   { /* get byte */
      I2C_SCLHIGH() ;
      data = (data << 1) | I2C_SDAGET() ;
      I2C_SCLLOW() ;
   }
   if (lastbyte)
      I2C_SDAHIGH() ; /* nack (high) */
   else
      I2C_SDALOW() ; /* ack (low) */
   I2C_SCLHIGH() ;
   I2C_SCLLOW() ;
   I2C_SDAHIGH() ; /* release pull down */

   return(data) ;
}

#endif


