﻿/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|             File Name: Modbus.h                                                                                       |
|                Author: Michael Krämer                                                                                 |
|                 Email: Michael@dk8pp.de                                                                               |
|                  Date: 21.01.2021                                                                                     |
|                  Tool: AVR Studio 7.0                                                                                 |
|                Device: ATTiny, ATMEGA, ATXMEGA                                                                        |
|               Version: 0.1                                                                                            |
|        Program status: still testing                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: include file for Modbus functions                                                                           |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#ifndef MODBUS

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Includes                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#define MODBUS 1

#define T_IMMEDIATE 0x80        /* immediate value in register table */
#define T_DATAPTR   0x40        /* pointer to data area in RAM */
#define T_EEPPTR    0x20        /* pointer to EEPROM */
#define T_MCBPTR    0x10        /* pointer to MCB in EEPROM */
#define T_PROGPTR   0x08        /* pointer to Program memory (Flash) */
#define T_EXECUTE   0x04        /* pointer to program code, execute it */
#define T_WORD      0x02        /* 16-bit word data */
#define T_RO        0x01        /* data is read-only */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       structs and enums                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/

typedef struct {
        uint16_t addr ;         /* modbus address */
        uint16_t count ;        /* number of elements */
        uint8_t type ;          /* type of register */
        uint16_t value ;        /* direct value, pointer to value or physical address in case of EEPROM or PROGMEM */
        uint16_t (*read) (uint16_t) ;  /* optional function for modbus read operation */
        uint16_t (*write) (uint16_t,uint16_t) ; /* optional function for modbus write operation, shall return true if successful */
        } MODBUS_REGISTER ;

typedef struct {
        MODBUS_REGISTER *ireg ; /* input registers */
        uint16_t ireg_count ;   /* number of entries in input register table */
        MODBUS_REGISTER *hreg ; /* holding registers */
        uint16_t hreg_count ;   /* number of entries in holding register table */
        MODBUS_REGISTER *coil ; /* coils */
        uint16_t coil_count ;   /* number of entries in coils table */
        MODBUS_REGISTER *file ; /* file descriptor */
        uint16_t file_count ;   /* number of entries in file records table */
        } MODBUS_DESCRIPTOR ;

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Function prototypes                                                                                             |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void modbus_message_handler(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc) ;
void modbus_error_response(TXBUFFER *txbuf, uint8_t errorcode) ;
uint16_t modbus_crc16(uint8_t *puchMsg, uint16_t usDataLen) ;
void modbus_message_received(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc) ;

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Inline functions                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       global variables                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/
_EXTERN_ bool broadcast ;
_EXTERN_ bool listen_only ; /* listen only mode if true */
_EXTERN_ bool restart_communications ; /* restart communication request received */
_EXTERN_ uint8_t module_address ; /* modbus address of this module */

/* NOINIT: the following variables survive the startup code */
/* Module status and error counters. May be read and written via ModBus functions */
_EXTERN_ uint16_t total_message_count_h __attribute__ ((section (".noinit"))) ; /* Total number of bus messages since last power-on reset (higher) */
_EXTERN_ uint16_t total_message_count_l __attribute__ ((section (".noinit"))) ; /* Total number of bus messages since last power-on reset (lower) */
_EXTERN_ uint16_t message_count_h __attribute__ ((section (".noinit"))) ; /* Number of bus messages for this module since last power-on reset (higher) */
_EXTERN_ uint16_t message_count_l __attribute__ ((section (".noinit"))) ; /* Number of bus messages for this module since last power-on reset (lower) */
_EXTERN_ uint16_t crc_error_count_h __attribute__ ((section (".noinit"))) ; /* Number of CRC errors since last power-on reset (higher) */
_EXTERN_ uint16_t crc_error_count_l __attribute__ ((section (".noinit"))) ; /* Number of CRC errors since last power-on reset (lower) */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       PROGMEM constants                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/
extern char const PROGMEM Modbus_FileName[] ;       /* Filename */
extern char const PROGMEM Modbus_TimeStamp[] ; /* Sun Sep 16 01:03:52 1973 */


#endif
